/*-
 * Copyright (c) 2003 Poul-Henning Kamp
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the authors may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <paths.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libgeom.h>

#include <sys/mman.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/devicestat.h>

#ifdef COMPAT_32BIT  
#define DEVCPY cpy_devstat64_to_32
#else
#define DEVCPY memcpy
#endif
/************************************************************/
static uint npages, pagesize, spp;
static int statsfd = -1;
static u_char *statp;
#ifdef COMPAT_32BIT
static int ndevstat64;
#endif
void
geom_stats_close(void)
{
	if (statsfd == -1)
		return;
	munmap(statp, npages *pagesize);
	statp = NULL;
	close (statsfd);
	statsfd = -1;
}

#ifdef COMPAT_32BIT
/* Takes the ceiling of a / b */
static inline long int_ceil(long a, long b)
{
	/*
	 * If it rounds exactly (0) then no change for ceiling
	 * Otherwise, int division will take the floor, so we need to add one
	 */
	return a % b != 0 ? a / b + 1 : a / b;
}
#endif

void
geom_stats_resync(void)
{
	void *p;

	if (statsfd == -1)
		return;
	for (;;) {
		p = mmap(statp, (npages + 1) * pagesize, 
		    PROT_READ, 0, statsfd, 0);
		if (p == MAP_FAILED)
			break;
		else
			statp = p;
		npages++;
	}
#ifdef COMPAT_32BIT
	ndevstat64 = npages * (pagesize / sizeof (struct devstat64));
	npages = int_ceil(npages * sizeof (struct devstat), sizeof (struct devstat64));
#endif
}

int
geom_stats_open(void)
{
	int error;
	void *p;

	if (statsfd != -1)
		return (EBUSY);
	statsfd = open(_PATH_DEV DEVSTAT_DEVICE_NAME, O_RDONLY);
	if (statsfd < 0)
		return (errno);
	pagesize = getpagesize();
	spp = pagesize / sizeof(struct devstat);
	p = mmap(NULL, pagesize, PROT_READ, 0, statsfd, 0);
	if (p == MAP_FAILED) {
		error = errno;
		close(statsfd);
		statsfd = -1;
		errno = error;
		return (error);
	}
	statp = p;
	npages = 1;
	geom_stats_resync();
	return (0);
}

struct snapshot {
	u_char		*ptr;
	uint		pages;
	uint		pagesize;
	uint		perpage;
	struct timespec	time;
	/* used by getnext: */
	uint		u, v;
};

#ifdef COMPAT_32BIT
static inline void bintime64_to_32(struct bintime *dst, const struct bintime64 *src)
{
	dst->sec = (u_long)(src->sec & 0xFFFFFFFF);
	dst->frac = src->frac;
}
static void devstat64_to_32(struct devstat *dst, const struct devstat64* src)
{
	int i;

	dst->sequence0 = src->sequence0;
	dst->allocated = src->allocated;
	dst->start_count = src->start_count;
	dst->end_count = src->end_count;
	bintime64_to_32(&dst->busy_from, &src->busy_from);
	dst->dev_links.stqe_next = src->dev_links.stqe_next;
	dst->device_number = src->device_number;
	memcpy(dst->device_name, src->device_name, DEVSTAT_NAME_LEN);
	dst->unit_number = src->unit_number;
	memcpy(dst->bytes,src->bytes, DEVSTAT_N_TRANS_FLAGS * sizeof (u_int64_t));
	memcpy(dst->operations, src->operations, DEVSTAT_N_TRANS_FLAGS * sizeof (u_int64_t));
	for(i = 0; i < DEVSTAT_N_TRANS_FLAGS; ++i)
	{
		bintime64_to_32(dst->duration + i, src->duration + i);
	}
	bintime64_to_32(&dst->busy_time, &src->busy_time);
	bintime64_to_32(&dst->creation_time, &src->creation_time);
	dst->block_size = src->block_size;
	memcpy(dst->tag_types, src->tag_types, 3 * sizeof (u_int64_t));
	dst->flags = src->flags;
	dst->device_type = src->device_type;
	dst->priority = src->priority;
	dst->id = src->id;
	dst->sequence1 = src->sequence1;
}
static void *cpy_devstat64_to_32(void *dst, const void *src, size_t len)
{
	struct devstat64 *cur64;
	struct devstat  *cur;
	int i;
	int spp64;
	int idx,idx64;
    
	idx = idx64 = 0;
	spp64 = pagesize / sizeof (struct devstat64);
	/* Loop through each devstat64 and assign it to a 32bit devstat
	 * This takes into account the fact that the kernel has the struct arrays
	 * Page aligned, so the last pagesize % sizeof(devstat64) bytes will be garbage.
	 */
	for(i = 0; i < ndevstat64; ++i)
	{
		idx = (i / spp) * pagesize + (i % spp) * sizeof (struct devstat);
		idx64 = (i / spp64) * pagesize + (i % spp64) * sizeof (struct devstat64);
		cur64 = (struct devstat64 *)((long)src +  idx64);
		cur = (struct devstat *)((long)dst +  idx);
	
		devstat64_to_32(cur, cur64);
	}
	return dst;
}
#endif
void *
geom_stats_snapshot_get(void)
{
	struct snapshot *sp;

	sp = malloc(sizeof *sp);
	if (sp == NULL)
		return (NULL);
	memset(sp, 0, sizeof *sp);
	sp->ptr = malloc(pagesize * npages);
	if (sp->ptr == NULL) {
		free(sp);
		return (NULL);
	}
	memset(sp->ptr, 0, pagesize * npages); 	/* page in, cache */
	clock_gettime(CLOCK_REALTIME, &sp->time);
	memset(sp->ptr, 0, pagesize * npages); 	/* page in, cache */
	DEVCPY(sp->ptr, statp,  pagesize * npages);
	sp->pages = npages;
	sp->perpage = spp;
	sp->pagesize = pagesize;
	return (sp);
}

void
geom_stats_snapshot_free(void *arg)
{
	struct snapshot *sp;

	sp = arg;
	free(sp->ptr);
	free(sp);
}

void
geom_stats_snapshot_timestamp(void *arg, struct timespec *tp)
{
	struct snapshot *sp;

	sp = arg;
	*tp = sp->time;
}

void
geom_stats_snapshot_reset(void *arg)
{
	struct snapshot *sp;

	sp = arg;
	sp->u = sp->v = 0;
}

struct devstat *
geom_stats_snapshot_next(void *arg)
{
	struct devstat *gsp;
	struct snapshot *sp;

	sp = arg;
	gsp = (struct devstat *)
	    (sp->ptr + sp->u * pagesize + sp->v * sizeof *gsp);
	if (++sp->v >= sp->perpage) {
		if (++sp->u >= sp->pages)
			return (NULL);
		else
			sp->v = 0;
	}
	return (gsp);
}
