#!/bin/sh
#
# installscript.sh
# Copyright 2004-2006 Sandvine Inc.
#
# This script is run on the test pc to installworld and other userland stuff.
# It expects that the kernel has been installed already as part of the build
# process.
#
# The following variables need to be set in this script's environment:
#
# TOPDIR                top mount point (one dir up from src/)
# svSrcFileBase         subdir of $TOPDIR with Sandvine source
# targetVersionStr
# CONF                  kernel conf file (e.g. STOCK)
# ARCH			i386/amd64

InstallFile()
{
    if [ -n "$1" ]
    then
        fileBase=$TOPDIR/$1
    else
        echo "No longer supported"
        exit 1
    fi
    
    install -o $4 -g $5 -m $6 $1 $2
}

createFile()
{
    local file=$1
    local owner=$2
    local group=$3
    local mode=$4

    if [ -f $file -a -z "$FORCE_UPDATE" ]; then
        return
    fi

    cat > $file
    chown $owner:$group $file
    chmod $mode $file
}

# Bail if anything goes wrong so we don't need to check every command.
# Note that -e doesn't exit on compound commands with && or ||.
set -e

# Force overwrite of files.
FORCE_UPDATE=1

# Check environment variables.
varsnotset=0
for var in \
    ARCH		\
    TOPDIR              \
    svSrcFileBase       \
    targetVersionStr    \
    CONF                ; do
    # can't use eval with FreeBSD sh or script exits due to set -e
    if sh -c "test -z \"\$$var\""; then
        echo "Environment variable \$$var not set"
        varsnotset=1
    fi
done
if [ $varsnotset != 0 ]; then
    exit 1
fi

if [ ! -c /dev/null ]; then
    echo "/dev/null does not exist -- mount devfs?"
    exit 1
fi

# Disklabel replaces /usr/local, /var/crash and /var/tmp with symlinks to
# space on d2.  If this is the case, move the originals back into the root
# filesystem so that rembo will pick up the correct files in a single fs.
if [ -h /usr/local -a -d /usr/local.old ]; then
    # Labify added socket and sudo packages after disklabel ran, so we need to
    # get rid of them first.  Labify will add them back on reboot.
    if [ -x /usr/local/bin/socket -a ! -x /usr/local.old/bin/socket ]; then
        pkg_delete 'socket-*'
    fi
    if [ -x /usr/local/bin/sudo -a ! -x /usr/local.old/bin/sudo ]; then
        pkg_delete 'sudo-*'
    fi
    rm -f /usr/local
    mv /usr/local.old /usr/local
fi
if [ -h /var/crash -a -d /var/crash.old ]; then
    rm -f /var/crash
    mv /var/crash.old /var/crash
fi
if [ -h /var/tmp ]; then
    rm -f /var/tmp
    mkdir -p /var/tmp
    chmod u+t,a+rwx /var/tmp
fi

# FreeBSD world.
# First Turn off schg flag for binaries, so that any obsolete files can be
# removed using the packing list later on.
chflags -R noschg /bin /sbin /usr/bin /usr/sbin /lib /usr/lib || true
# Some files are installed only if they don't already exist.  Remove them here
# so we always get the one out of clearcase.
rm -f /boot/loader.rc
# The target dir /usr/share/mk files get used so update them first.  This is
# kind of a hack and should be reconsidered or possibly fixed in another way.
cp /usr/src/share/mk/* /usr/share/mk
# Now do the actual installworld.
cd /usr/src
env STRIPBIN=/usr/src/sv/stripbin.sh make installworld

# Boot config (serial console).
createFile /boot.config root wheel 644 <<EOF
-h
EOF
# Root's config files.
createFile /root/.bashrc root wheel 644 <<EOF
PS1="# "
alias ls="ls -F"
set -o vi
if [ -f /etc/profile ]
then
    . /etc/profile
fi
EOF
createFile /root/.inputrc root wheel 644 <<EOF
set editing-mode vi
EOF
# Loader configuration
createFile /boot/loader.conf root wheel 644 <<EOF
# Please make all changes to /boot/loader.conf.local
# and not this file

kern.imgver=7.00.0000

userconfig_script_load="YES"
autoboot_delay=4
console=comconsole

kern.timecounter.smp_tsc=1
debug.mpsafevfs=0

watchdog_load="YES"
rasum_load="YES"
if_vlan_load="YES"
hw.ahd0.max_tags=32
boot_default="yes"      # starts boot with -r option to use ROOTDEVNAME
beastie_disable="YES"

# Limits for ia32 apps on amd64 kernels
compat.ia32.maxdsiz="1610612736"	# 1.5 GB

# Device driver tuning
hw.em.txd="4096"
hw.em.rxd="4096"

# Stack tuning
net.inet.ip.intr_queue_maxlen="4096"

# \$auto_kernel is set by loader.rc for SMP or UP mode.
bootfile="\${auto_kernel}"
EOF
createFile /boot/loader.conf.local root wheel 644 <<EOF
# Please make changes to this file, and not /boot/loader.conf
EOF
createFile /etc/rc.conf.local root wheel 644 <<EOF
# Please make changes to this file, and not /etc/rc.conf
EOF
createFile /etc/sysctl.conf.local root wheel 644 <<EOF
# Please make changes to this file, and not /etc/sysctl.conf
EOF
createFile /etc/periodic.conf root wheel 644 <<EOF
# Please make all changes to /etc/periodic.conf.local
# and not this file.

daily_show_success="NO"
weekly_show_success="NO"
monthly_show_success="NO"
security_show_success="NO"

daily_output="/var/log/daily.log"
weekly_output="/var/log/weekly.log"
monthly_output="/var/log/monthly.log"
# Since we're only showing interesting output anyways, just append
# security output to the daily.log
daily_status_security_output=""

daily_clean_tmps_enable="YES"
daily_clean_tmps_dirs="/tmp /d2/tmp"
daily_status_mailq_shorten="YES"

daily_security_checksetuid_enable="NO"
daily_clean_disks_enable="NO"
daily_accounting_enable="NO"
daily_news_expire_enable="NO"

daily_status_disks_enable="NO"
daily_status_network_enable="NO"
daily_status_rwho_enable="NO"
daily_status_named_enable="NO"

weekly_locate_enable="NO"
weekly_whatis_enable="NO"

monthly_accounting_enable="NO"
EOF
createFile /etc/periodic.conf.local root wheel 644 <<EOF
# Please make changes to this file, and not /etc/periodic.conf
EOF
createFile /etc/sv_group root wheel 600 <<EOF
sv_service	sv_operator,sv_service
sv_admin	sv_operator,sv_service,sv_admin
EOF
: | createFile /etc/sv_group.local root wheel 600
createFile /etc/devfs.rules root wheel 444 <<EOF
# Please make all changes to /etc/devfs.rules.local, not this file.

# Allows read/write access to ttyd1
[devfsrules_perms=100]
add path 'ttyd1' mode 666
EOF
createFile /etc/devfs.rules.local root wheel 444 <<EOF
# Please makes changes to this file, and not /etc/devfs.rules
EOF
createFile /etc/ntp.conf root wheel 644 <<EOF
server 0.pool.ntp.org
server 1.pool.ntp.org
server 2.pool.ntp.org

driftfile /etc/ntp.drift
EOF

install -o root -g wheel -m 755 /usr/src/sv/misc/fetchlabify /etc/rc.d

# After "make distribution" the password file is the one out of clearcase,
# not the one provided by labify.  SSH to this pc will not work any longer.
cd /usr/src/etc
make DESTDIR=/ distribution

# Standard FreeBSD target - kernel.
cd /usr/src
make KERNCONF=$CONF installkernel
install -p -m 755 -o root -g wheel \
    /usr/obj/usr/src/sys/$CONF/kernel.debug /var/crash/kernel.debug
rm -f /var/crash/kernel.debug.gz
gzip /var/crash/kernel.debug

# Uniprocessor kernel.
if [ -f /usr/src/sys/i386/conf/$CONF-NOSMP ]; then
    target=/boot/kernel
    install -p -m 555 -o root -g wheel /usr/obj/usr/src/sys/$CONF-NOSMP/kernel $target/kernel.NOSMP
    install -p -m 755 -o root -g wheel /usr/obj/usr/src/sys/$CONF-NOSMP/kernel.debug /var/crash/kernel.NOSMP.debug
    rm -f /var/crash/kernel.NOSMP.debug.gz
    gzip -f /var/crash/kernel.NOSMP.debug
fi

# Debug kernel.
debugkernel=/usr/obj/usr/src/sys/SPB-DEBUG/kernel.debug
if [ -f $debugkernel ]; then
    target=/boot/kernel
    install -p -m 755 -o root -g wheel $debugkernel $target/debugkernel
fi

# Store debug modules under /var/crash/.
for CONF in STOCK STOCK-NOSMP DEBUG ; do
    moddir=/usr/obj/usr/src/sys/$CONF
    if [ ! -f $moddir ] ; then
	continue
    fi
    mkdir -p /var/crash/$CONF
    find $moddir/ -name '*.ko.debug' \
	-exec cp {} /var/crash/$CONF \;
done

# Create empty log files
/usr/sbin/newsyslog -NCC

# Mount points, empty dirs for installs
mkdir -p /d2 /usr/compat
ln -fs /usr/compat /compat

# SV user creation
for user in sv sv_operator sv_service sv_admin; do
    homedir=/home/$user
    mkdir -p $homedir
    chmod 700 $homedir
    chown $user:$user $homedir

    if [ $user = sv ]; then
        continue
    fi
    for file in .bashrc .cshrc .inputrc .k5login .login .profile; do
        cp -f /root/$file $homedir/
        chown $user:$user $homedir/$file
    done
    tFile=`mktemp -t profile_update`
    sed "s|^HOME=.*\$|HOME=${homedir}|" $homedir/.profile > $tFile
    cat $tFile > $homedir/.profile
    rm $tFile
done

# reqst00064096 Run quickstart on first sv_admin login
cat >>/home/sv_admin/.profile <<EOF
if [ ! -f /etc/quickstart-done -a -f /usr/local/sandvine/bin/quickstart ]; then
	sudo /usr/local/sandvine/bin/quickstart -i
fi
EOF

# XXX Gross, build once inside the chroot, then install only, after
# we're mounted on the TPC (since we can't write as root over NFS).
domake=0
if [ ! -f $TOPDIR/$svSrcFileBase/rasum/rasum.o ] ; then
    domake=1
fi

# Sandvine modules.
dirlist="rasum sysinfo watchdog"
for dir in $dirlist; do
    cd $TOPDIR/$svSrcFileBase/$dir
    [ $domake -ne 0 ] && make
    make install
    if [ "$dir" = watchdog ]; then
        make install_wddt
    fi
done

case "$ARCH" in
i386)
    packagedir=/usr/src/sv/packages
    old_i386_lib_dir=/lib
    old_i386_usr_lib_dir=/usr/lib
    old_i386_usr_lib_compat_dir=/usr/lib/compat

    cd $TOPDIR/$svSrcFileBase/memtest86_plus
    [ $domake -ne 0 ] && gmake
    make install-bsd
    ;;
amd64)
    packagedir=/usr/src/sv/packages/amd64-FreeBSD6
    old_i386_lib_dir=/usr/lib32
    old_i386_usr_lib_dir=/usr/lib32
    old_i386_usr_lib_compat_dir=/usr/lib32
    ;;
esac

# reqst00058585 use libthr instead of libpthread
createFile /etc/libmap.conf root wheel 644 <<EOF
[/usr/local/diablo-jdk1.5.0/bin/]
libpthread.so.2		libthr.so.2

[/d2/local/diablo-jdk1.5.0/bin/]
libpthread.so.2		libthr.so.2
EOF

# Sandvine files.
mkdir -p /var/db/copyrights
mkdir -p /etc/sv/mobomap
mkdir -p /usr/local/libdata

# /usr/local/libdata/image_packing_list is now set by the Makefile packing_list
# target, which will do the glob expansion and warn for missing files
#install -o root -g wheel -m 644 \
#    $svSrcFileBase/maintenance_scripts/image_packing_list /usr/local/libdata
install -o root -g wheel -m 444 \
    $svSrcFileBase/maintenance_scripts/libmtrie_copyright.txt /var/db/copyrights
install -o root -g wheel -m 444 \
    $svSrcFileBase/maintenance_scripts/pcre_copyright.txt /var/db/copyrights
install -o root -g wheel -m 444 \
    $svSrcFileBase/maintenance_scripts/umac_copyright.txt /var/db/copyrights
install -o root -g sv_service -m 750 \
    $svSrcFileBase/maintenance_scripts/0000.bypass.sh /usr/local/etc/rc.d
install -o root -g sv_service -m 750 \
    $svSrcFileBase/maintenance_scripts/0000.disklabel.sh /etc/rc.d
install -o root -g sv_service -m 750 \
    $svSrcFileBase/maintenance_scripts/0000.privilege.sh /usr/local/etc/rc.d
install -o root -g wheel -m 755 \
    $svSrcFileBase/maintenance_scripts/ipfwstat /usr/local/bin
install -o root -g wheel -m 755 \
    $svSrcFileBase/maintenance_scripts/410.sv-status-pkg /etc/periodic/weekly
    
cp -r $svSrcFileBase/mobomap/* /etc/sv/mobomap
for file in `find /etc/sv/mobomap -type d` ; do
    chmod 755 $file
done
for file in `find /etc/sv/mobomap -type f` ; do
    chmod 754 $file
done
    
OLD_I386_LIBS="
libbsdxml.so.1
libc.so.5
libcrypt.so.2
libcrypto.so.3
libm.so.3
libncurses.so.5
libreadline.so.5
libutil.so.4
libz.so.2
"

OLD_I386_USR_LIBS="
libbz2.so.1
libpthread.so.1
libssl.so.3
libstdc++.so.4
"

OLD_I386_USR_LIB_COMPATS="
libm.so.2
libstdc++.so.3
"

for lib in $OLD_I386_LIBS; do
    install -o root -g wheel -m 444 \
	/usr/src/sv/lib/i386/$lib $old_i386_lib_dir
done

for lib in $OLD_I386_USR_LIBS; do
    install -o root -g wheel -m 444 \
	/usr/src/sv/lib/i386/$lib $old_i386_usr_lib_dir
done

for lib in $OLD_I386_USR_LIB_COMPATS; do
    install -o root -g wheel -m 444 \
	/usr/src/sv/lib/i386/$lib $old_i386_usr_lib_compat_dir
done

# reqst00024778: add ftpd-nologin shell for ftp only access
ftpdnologin=/usr/local/sbin/ftpd-nologin
if [ ! -f $ftpdnologin ]; then
    nologin=`which nologin`
    if [ -z "$nologin" ]; then
        echo "Required binary 'nologin' not found"
        exit 1
    fi
    ln -fs $nologin $ftpdnologin
fi
if ! grep -q "$ftpdnologin" /etc/shells; then
    echo "$ftpdnologin" >> /etc/shells
fi

# prep-image should be called by this script (or by the overall build script)
# so some of the following steps can be removed.  For now this is a minimal
# set of operations needed so that the image comes up correctly the next time.

truncate -s 0 /etc/make.conf
awk "BEGIN { print \"kern.imgver=$targetVersionStr\" } 
     ! /^kern.imgver/ { print }" /boot/loader.conf > /tmp/loader.conf.tmp
mv /tmp/loader.conf.tmp /boot/loader.conf
chmod 644 /boot/loader.conf
chown root:wheel /boot/loader.conf

# Add / upgrade bash
pkg_info | awk '/^bash/ { print $1 }' | xargs pkg_delete
pkg_add $packagedir/bash-[23]*.t[bg]z
# bash_completion now built from src/usr.local/etc/bash_completion. bash should
# be done the same way as well
#pkg_add $packagedir/bash-completion*.t[bg]z
install /usr/local/bin/bash /bin/bash


# Create a virtual package containing the files we will keep on the image
# (excluding ones we expect to change) for pkg_info -g
/usr/src/sv/make_pseudo_pkg.sh \
    $TOPDIR/$svSrcFileBase/maintenance_scripts/image_packing_list \
    /usr/src/sv/pkg_exclude \
    svstockimage

# So that labify runs next time
rm -f /etc/labify-done
rm -f /var/db/dhclient.leases
rm -f /usr/local/etc/sudoers

# Labify backs up ssh config to .orig files if they don't already exist.
# Remove them here so it will back up our new ones, to be replaced by
# prep_image before imaging.
rm -f /etc/ssh/*.orig

