#!/bin/sh

# This script copies programs, directories and files needed to do a buildworld
# and buildkernel into a directory to prepare for a chrooted build.
#
# Some of these are used in the preliminary stages.  For example /usr/include
# is used to build an up-to-date make to carry on the build process (but that
# make does not end up in the installed world).  Surprisingly, some of these
# are used for files that end up in the image: /usr/share/misc/termcap is used
# when building usr/sbin/sysinstall.
#
# Each program in CHROOT_PROGS is located with which(1) and its shared library
# deps are found with ldd(1) and the results of both are copied into the dir.
# CHROOT_DIRS and CHROOT_FILES are copied over without any processing.

CHROOT_PROGS="
    [
    as ar awk
    basename btxld bzip2
    c++ cap_mkdb cat cc chflags chmod chown cmp colldef
    compile_et cp cpp crunchgen
    crunchide
    date dd diff dirname
    echo egrep env expand ex expr
    false fgrep file2c find
    gencat gensnmptree gmake grep groff gunzip gzip
    head hostname
    id install
    install-info
    join
    kbdcontrol
    ld lex 
    lockf
    lorder ln ls
    m4 make mkdep mkdir mklocale mktemp mount mount_devfs mtree mv
    nm
    paste patch printf pwd pwd_mkdb
    qemu-img
    ranlib rm rpcgen
    sed sh sort sysctl
    tail test touch tr true tsort
    umount uname unifdef uniq uudecode uuencode
    vgrind
    wc which whoami
    xargs
    yacc
    zic
    /usr/libexec/cc1
    /usr/libexec/cc1plus
    /usr/local/sbin/makefs
"
CHROOT_DIRS="
    /usr/include
    /usr/share/mk
"

CHROOT_FILES="
    /usr/lib/crt1.o
    /usr/lib/crti.o
    /usr/lib/crtn.o
    /usr/lib/crtbegin.o
    /usr/lib/crtend.o
    /usr/lib/libgcc.a
    /usr/lib/libc.a
    /usr/lib/libm.a
    /usr/lib/libl.a
    /usr/lib/libz.a
    /usr/lib/libtermcap.a
    /usr/lib/libstdc++.a
    /libexec/ld-elf.so.1
    /usr/libexec/vfontedpr
    /usr/share/tmac/tmac.vgrind
    /usr/share/misc/termcap
    /etc/passwd
    /etc/pwd.db
    /etc/group
"

# $1 - Top dir to become root (/)
setupChroot()
{
    local topdir=$1
    local files
    files="$CHROOT_FILES"

    # Copy entire directories
    for directory in $CHROOT_DIRS; do
        mkdir -p $topdir/$directory
        cp -f -p -R $directory $topdir/`dirname $directory`
    done

    # Find actual programs and their shared lib deps
    for prog in $CHROOT_PROGS; do
	if ! fullpath=$(which $prog); then
	    echo "$prog not found"
	    exit 1
	fi
        shlibs=$(ldd $fullpath 2>/dev/null | sed -n 's/.* => \(.*\) .*$/\1/p')
        files="$files $fullpath $shlibs"
    done

    # Copy the list of files into the chroot
    for file in $files; do
        dir=$(dirname $file)
        mkdir -p $topdir/$dir
        chflags noschg $topdir/$file 2>/dev/null || true
        cp -fp $file $topdir/$file 
    done

    mkdir -p $topdir/dev
    chmod 555 $topdir/dev
    mkdir -p $topdir/tmp
    chmod 1777 $topdir/tmp
}

# Bail if anything goes wrong
set -e

# Simple sanity checks on the chosen directory
if [ -z "$1" ]; then
    echo "usage: `basename $0` directory"
    exit 1
fi
if [ "$1" = "/" ]; then
    echo "not copying into /" 
    exit 1
fi
if [ ! -d "$1" ]; then
    echo "directory $1 does not exist"
    exit 1
fi

if [ "$(id -u)" != 0 ]; then
    echo "$(basename $0) must be run as root."
    exit 1
fi

setupChroot $1
