#!/bin/sh
#
# installscript.sh
# Copyright 2004 Sandvine Inc.
#
# This script is run on the test pc to installworld and other userland stuff.
# It expects that the kernel has been installed already as part of the build
# process.
#
# The following variables need to be set in this script's environment:
#
# TOPDIR                top mount point (one dir up from src/)
# svSrcFileBase         subdir of $TOPDIR with Sandvine source
# targetVersionStr
# CONF                  kernel conf file (e.g. STOCK)

InstallFile()
{
    if [ -n "$1" ]
    then
        fileBase=$TOPDIR/$1
    else
        echo "No longer supported"
        exit 1
    fi
    
    install -o $4 -g $5 -m $6 $fileBase/$2 $3
}

createFile()
{
    local file=$1
    local owner=$2
    local group=$3
    local mode=$4

    if [ -f $file -a -z "$FORCE_UPDATE" ]; then
        return
    fi

    cat > $file
    chown $owner:$group $file
    chmod $mode $file
}

# Bail if anything goes wrong so we don't need to check every command.
# Note that -e doesn't exit on compound commands with && or ||.
set -e

# Force overwrite of files.
FORCE_UPDATE=1

# Check environment variables.
varsnotset=0
for var in \
    TOPDIR              \
    svSrcFileBase       \
    targetVersionStr    \
    CONF                ; do
    # can't use eval with FreeBSD sh or script exits due to set -e
    if sh -c "test -z \"\$$var\""; then
        echo "Environment variable \$$var not set"
        varsnotset=1
    fi
done
if [ $varsnotset != 0 ]; then
    exit 1
fi

if [ ! -c /dev/null ]; then
    echo "/dev/null does not exist -- mount devfs?"
    exit 1
fi

# Disklabel replaces /usr/local, /var/crash and /var/tmp with symlinks to
# space on d2.  If this is the case, move the originals back into the root
# filesystem so that rembo will pick up the correct files in a single fs.
if [ -h /usr/local -a -d /usr/local.old ]; then
    # Labify added socket and sudo packages after disklabel ran, so we need to
    # get rid of them first.  Labify will add them back on reboot.
    if [ -x /usr/local/bin/socket -a ! -x /usr/local.old/bin/socket ]; then
        pkg_delete 'socket-*'
    fi
    if [ -x /usr/local/bin/sudo -a ! -x /usr/local.old/bin/sudo ]; then
        pkg_delete 'sudo-*'
    fi
    rm -f /usr/local
    mv /usr/local.old /usr/local
fi
if [ -h /var/crash -a -d /var/crash.old ]; then
    rm -f /var/crash
    mv /var/crash.old /var/crash
fi
if [ -h /var/tmp ]; then
    rm -f /var/tmp
    mkdir -p /var/tmp
    chmod u+t,a+rwx /var/tmp
fi

# FreeBSD world.
# First Turn off schg flag for binaries, so that any obsolete files can be
# removed using the packing list later on.
chflags -R noschg /bin /sbin /usr/bin /usr/sbin /lib /usr/lib || true
# Some files are installed only if they don't already exist.  Remove them here
# so we always get the one out of clearcase.
rm -f /boot/loader.rc
# The target dir /usr/share/mk files get used so update them first.  This is
# kind of a hack and should be reconsidered or possibly fixed in another way.
cp /usr/src/share/mk/* /usr/share/mk
# Now do the actual installworld.
cd /usr/src
make installworld

## Boot config (serial console).
#createFile /boot.config root wheel 644 <<EOF
#-h
#EOF
# Root's config files.
createFile /root/.bashrc root wheel 644 <<EOF
PS1="# "
alias ls="ls -F"
set -o vi
if [ -f /etc/profile ]
then
    . /etc/profile
fi
EOF
createFile /root/.inputrc root wheel 644 <<EOF
set editing-mode vi
EOF
# Loader configuration
createFile /boot/loader.conf root wheel 644 <<EOF
# Please make all changes to /boot/loader.conf.local
# and not this file

kern.imgver=7.00.0000

userconfig_script_load="YES"
autoboot_delay=4
console=vidconsole

kern.timecounter.smp_tsc=1

watchdog_load="YES"
rasum_load="YES"
if_vlan_load="YES"
hw.ahd0.max_tags=32
mce_load="YES"
usb_load="YES"
umass_load="YES"
boot_default="yes"      # starts boot with -r option to use ROOTDEVNAME
beastie_disable="YES"

# \$auto_kernel is set by loader.rc for SMP or UP mode.
bootfile="\${auto_kernel}"
vfs.root.mountfrom="ufs:/dev/da0s1a"
EOF
createFile /boot/loader.conf.local root wheel 644 <<EOF
# Please makes changes to this file, and not /boot/loader.conf
EOF
createFile /etc/rc.conf.local root wheel 644 <<EOF
# Please makes changes to this file, and not /etc/rc.conf
EOF
createFile /etc/devfs.rules root wheel 444 <<EOF
# Please make all changes to /etc/devfs.rules.local, not this file.

# Allows read/write access to ttyd1 and bpf devices
[devfsrules_perms=100]
add path 'ttyd1' mode 666
add path 'bpf*' mode 666
EOF
createFile /etc/devfs.rules.local root wheel 444 <<EOF
# Please makes changes to this file, and not /etc/devfs.rules
EOF

createFile /etc/fstab root wheel 444 <<EOF
/dev/da0s1a    /         ufs    ro         1  0
md             /tmp      mfs    rw,-s32m   2  0
md             /var/tmp  mfs    rw,-s10m   3  0
md             /var/log  mfs    rw,-s10m   4  0
EOF

createFile /etc/ttys root wheel 444 <<EOF
console none                            unknown off insecure
#
ttyv0   "/usr/libexec/getty Pc"         cons25  on  secure
# Virtual terminals
ttyv1   "/usr/libexec/getty Pc"         cons25  on  secure
ttyv2   "/usr/libexec/getty Pc"         cons25  on  secure
ttyv3   "/usr/libexec/getty Pc"         cons25  on  secure
ttyv4   "/usr/libexec/getty Pc"         cons25  on  secure
ttyv5   "/usr/libexec/getty Pc"         cons25  on  secure
# Serial terminals
# The 'dialup' keyword identifies dialin lines to login, fingerd etc.
ttyd0   "/usr/libexec/getty std.9600"   dialup  off secure
ttyd1   "/usr/libexec/getty std.115200" vt100   on  secure
EOF

createFile /etc/gettytab root wheel 444 <<EOF
default:\
        :al=root:cb:ce:ck:lc:fd#1000:im=\r\n%s/%m (%h) (%t)\r\n\r\n:sp#1200:\
        :if=/etc/issue:

2|std.9600|9600-baud:\
        :np:sp#9600:
g|std.19200|19200-baud:\
        :np:sp#19200:
std.38400|38400-baud:\
        :np:sp#38400:
std.57600|57600-baud:\
        :np:sp#57600:
std.115200|115200-baud:\
        :np:sp#115200:
std.230400|230400-baud:\
        :np:sp#230400:

P|Pc|Pc console:\
        :ht:np:sp#115200:
EOF

# After "make distribution" the password file is the one out of clearcase,
# not the one provided by labify.  SSH to this pc will not work any longer.
cd /usr/src/etc
make distribution

# Standard FreeBSD target - kernel.
cd /usr/src
make KERNCONF=$CONF installkernel
install -p -m 755 -o root -g wheel \
    /usr/obj/usr/src/sys/$CONF/kernel.debug /var/crash/kernel.debug
rm -f /var/crash/kernel.debug.gz
gzip /var/crash/kernel.debug

# Uniprocessor kernel.
if [ -f /usr/src/sys/i386/conf/$CONF-NOSMP ]; then
    target=/boot/kernel
    install -p -m 555 -o root -g wheel /usr/obj/usr/src/sys/$CONF-NOSMP/kernel $target/kernel.NOSMP
    install -p -m 755 -o root -g wheel /usr/obj/usr/src/sys/$CONF-NOSMP/kernel.debug /var/crash/kernel.NOSMP.debug
    rm -f /var/crash/kernel.NOSMP.debug.gz
    gzip -f /var/crash/kernel.NOSMP.debug
fi

## Debug kernel.
#if [ -f /usr/src/sys/i386/conf/DEBUG ]; then
    #target=/boot/kernel
    #install -p -m 755 -o root -g wheel /usr/obj/usr/src/sys/DEBUG/kernel.debug $target/debugkernel
#fi

## Store debug modules under /var/crash/.
#for CONF in STOCK STOCK-NOSMP DEBUG ; do
    #mkdir -p /var/crash/$CONF
    #find /usr/obj/usr/src/sys/$CONF/ -name '*.ko.debug' \
	#-exec cp {} /var/crash/$CONF \;
#done

# SV user creation
for user in sv sv_operator sv_service sv_admin; do
    homedir=/home/$user
    mkdir -p $homedir
    chmod 700 $homedir
    chown $user:$user $homedir

    if [ $user = sv ]; then
        continue
    fi
    for file in .bashrc .cshrc .inputrc .k5login .login .profile; do
        cp -f /root/$file $homedir/
        chown $user:$user $homedir/$file
    done
    tFile=`mktemp -t profile_update`
    sed "s|^HOME=.*\$|HOME=${homedir}|" $homedir/.profile > $tFile
    cat $tFile > $homedir/.profile
    rm $tFile
done

createFile /root/bin/runmenu.sh root 777 << EOF
#!/bin/sh
/usr/local/sandvine/CET/runscript.sh /root/testmenu.tcl

EOF
pw usermod root -s /root/bin/runmenu.sh




# XXX Gross, build once inside the chroot, then install only, after
# we're mounted on the TPC (since we can't write as root over NFS).
domake=0
if [ ! -f $TOPDIR/$svSrcFileBase/lmmon/lmmon.o ] ; then
    domake=1
fi

# Sandvine modules.
for dir in lmmon rasum siocontrol sysinfo mce watchdog; do
    cd $TOPDIR/$svSrcFileBase/$dir
    [ $domake -ne 0 ] && make
    make install
    if [ "$dir" = watchdog ]; then
        make install_wddt
    fi
done
cd $TOPDIR/$svSrcFileBase/memtest86_plus
[ $domake -ne 0 ] && gmake
make install-bsd

# Sandvine files.
mkdir -p /var/db/copyrights
mkdir -p /usr/local/libdata

InstallFile $svSrcFileBase maintenance_scripts/libmtrie_copyright.txt /var/db/copyrights root wheel 444
InstallFile $svSrcFileBase maintenance_scripts/pcre_copyright.txt /var/db/copyrights root wheel 444
InstallFile $svSrcFileBase maintenance_scripts/umac_copyright.txt /var/db/copyrights root wheel 444
#InstallFile $svSrcFileBase maintenance_scripts/0000.bypass.sh /usr/local/etc/rc.d root sv_service 750
#InstallFile $svSrcFileBase maintenance_scripts/0000.disklabel.sh /usr/local/etc/rc.d root sv_service 750
#InstallFile $svSrcFileBase maintenance_scripts/0000.privilege.sh /usr/local/etc/rc.d root sv_service 750
InstallFile $svSrcFileBase maintenance_scripts/mk_inventory /usr/local/bin root wheel 544
InstallFile $svSrcFileBase maintenance_scripts/prep_image /usr/local/bin root wheel 544
InstallFile $svSrcFileBase maintenance_scripts/ipfwstat /usr/local/bin root wheel 755
InstallFile $svSrcFileBase maintenance_scripts/410.sv-status-pkg /etc/periodic/weekly root wheel 755

InstallFile $svSrcFileBase maintenance_scripts/image_packing_list /usr/local/libdata root wheel 644

# reqst00024778: add ftpd-nologin shell for ftp only access
ftpdnologin=/usr/local/sbin/ftpd-nologin
if [ ! -f $ftpdnologin ]; then
    nologin=`which nologin`
    if [ -z "$nologin" ]; then
        echo "Required binary 'nologin' not found"
        exit 1
    fi
    ln -fs $nologin $ftpdnologin
fi
if ! grep -q "$ftpdnologin" /etc/shells; then
    echo "$ftpdnologin" >> /etc/shells
fi

# prep-image should be called by this script (or by the overall build script)
# so some of the following steps can be removed.  For now this is a minimal
# set of operations needed so that the image comes up correctly the next time.

truncate -s 0 /etc/make.conf
awk "BEGIN { print \"kern.imgver=$targetVersionStr\" } 
     ! /^kern.imgver/ { print }" /boot/loader.conf > /tmp/loader.conf.tmp
mv /tmp/loader.conf.tmp /boot/loader.conf
chmod 644 /boot/loader.conf
chown root:wheel /boot/loader.conf

# Add / upgrade bash
pkg_info | awk '/^bash/ { print $1 }' | xargs pkg_delete
packagedir=/usr/src/sv/packages
pkg_add $packagedir/bash-[23]*.t[bg]z
pkg_add $packagedir/bash-completion*.t[bg]z
pkg_add $packagedir/socket*.t[bg]z

#TODO add these packages
#pkg_add tclkit
#pkg_add svcet
# setup autologin
# remove unneded startup scripts


install /usr/local/bin/bash /bin/bash


## Create a virtual package containing the files we will keep on the image
## (excluding ones we expect to change) for pkg_info -g
#/usr/src/sv/make_pseudo_pkg.sh \
#    $TOPDIR/$svSrcFileBase/maintenance_scripts/image_packing_list \
#    /usr/src/sv/pkg_exclude \
#    svstockimage

# So that labify runs next time
rm -f /etc/labify-done
rm -f /var/db/dhclient.leases
rm -f /usr/local/etc/sudoers

# Labify backs up ssh config to .orig files if they don't already exist.
# Remove them here so it will back up our new ones, to be replaced by
# prep_image before imaging.
rm -f /etc/ssh/*.orig

# Done, root directory ready to be tar-ed up
