/*-
 * Copyright (c) 2005 Sandvine Incorporated.  All righs reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Ed Maste <emaste@phaedrus.sandvine.ca>
 */

/*
 * This module detects Intel Multiprocessor spec info (mptable) and exports
 * into the environment.
 *
 * This is intended for use in large deployments consisting of both
 * uniprocessor and SMP machines, sharing a common disk image or PXE boot
 * configuration.  With a little custom Forth the loader can automatically
 * choose an apropriate kernel.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <stand.h>
#include <btxv86.h>
#include <mptable.h>

#define MPFPS_SIG "_MP_"
#define MPCTH_SIG "PCMP"

static mpfps_t biosmptable_find_mpfps(void);
static mpfps_t biosmptable_search_mpfps(char *base, int length);
static mpcth_t biosmptable_check_mpcth(char *addr);

void
biosmptable_detect(void)
{
    mpfps_t	mpfps;
    char	buf[16];
    int 	ncpu = 0;
    int		nioapic = 0;

    /* locate and validate the mpfps */
    if ((mpfps = biosmptable_find_mpfps()) == NULL)
	return;

    /* 
     * If thie config_type is nonzero then this is a default configuration
     * from Chapter 5 in the MP spec.  Report 2 cpus and 1 I/O APIC.
     */
    if (mpfps->config_type != 0) {
    	ncpu = 2;
	nioapic = 1;
    } else {
    	mpcth_t mpcth;
	char*	entry_type_p;
	int	i;
	
	mpcth = biosmptable_check_mpcth(PTOV(mpfps->pap));
	if (mpcth == NULL)
	    return;
	entry_type_p = (char *)(mpcth + 1);
	for (i = 0; i < mpcth->entry_count; i++) {
	    switch (*entry_type_p) {
	    case 0:
		entry_type_p += sizeof(struct PROCENTRY);
		ncpu++;
		break;
	    case 1:
		entry_type_p += sizeof(struct BUSENTRY);
		break;
	    case 2:
	    	entry_type_p += sizeof(struct IOAPICENTRY);
		nioapic++;
		break;
	    case 3:
	    case 4:
	    	entry_type_p += sizeof(struct INTENTRY);
		break;
	    default:
	    	printf("mptable: unknown entry type (%d)\n", *entry_type_p);
		return;
	    }
	}
    }

    /* Export number of cpus and apics */
    sprintf(buf, "%d", ncpu);
    setenv("mptable.ncpu", buf, 1);
    sprintf(buf, "%d", nioapic);
    setenv("mptable.nioapic", buf, 1);
}

/*
 * Find the MP Floating Pointer Structure.  See the MP spec section 4.1.
 */
static mpfps_t
biosmptable_find_mpfps(void)
{
    mpfps_t	mpfps;
    uint16_t	*addr;

    /* EBDA is the 1 KB addressed by the 16 bit pointer at 0x40E. */
    addr = (uint16_t *)PTOV(0x40E);
    mpfps = biosmptable_search_mpfps(PTOV(*addr << 4), 0x400);
    if (mpfps != NULL)
	return (mpfps);

    /* Check the BIOS. */
    mpfps = biosmptable_search_mpfps(PTOV(0xf0000), 0x10000);
    if (mpfps != NULL)
	return (mpfps);

    return (NULL);
}

static mpfps_t
biosmptable_search_mpfps(char *base, int length)
{
    mpfps_t	mpfps;
    u_int8_t	*cp, sum;
    int		ofs, idx;

    /* search on 16-byte boundaries */
    for (ofs = 0; ofs < length; ofs += 16) {
	mpfps = (mpfps_t)(base + ofs);

	/* compare signature, validate checksum */
	if (!strncmp(mpfps->signature, MPFPS_SIG, strlen(MPFPS_SIG))) {
	    cp = (u_int8_t *)mpfps;
	    sum = 0;
	    /* mpfps is 16 bytes, or one "paragraph" */
	    if (mpfps->length != 1) {
	    	printf("mptable: bad mpfps length (%d)\n", mpfps->length);
		continue;
	    }
	    for (idx = 0; idx < mpfps->length * 16; idx++)
		sum += *(cp + idx);
	    if (sum != 0) {
		printf("mptable: bad mpfps checksum (%d)\n", sum);
		continue;
	    }
	    return(mpfps);
	}
    }
    return(NULL);
}

static mpcth_t
biosmptable_check_mpcth(char *addr)
{
    mpcth_t	mpcth;
    u_int8_t	*cp, sum;
    int		idx;

    /* mpcth must be in the first 1MB */
    if ((u_int32_t)addr >= 1024 * 1024) {
	printf("mptable: bad mpcth address (%p)\n", addr);
	return(NULL);
    }

    /* Compare signature and validate checksum. */
    mpcth = (mpcth_t)addr;
    if (strncmp(mpcth->signature, MPCTH_SIG, strlen(MPCTH_SIG)) != 0) {
        printf("mptable: bad mpcth signature\n");
	return(NULL);
    }
    cp = (u_int8_t *)mpcth;
    sum = 0;
    for (idx = 0; idx < mpcth->base_table_length; idx++)
	sum += *(cp + idx);
    if (sum != 0) {
	printf("mptable: bad mpcth checksum (%d)\n", sum);
	return(NULL);
    }

    return mpcth;
}

