/*-
 * Copyright 1996 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that both the above copyright notice and this
 * permission notice appear in all copies, that both the above
 * copyright notice and this permission notice appear in all
 * supporting documentation, and that the name of M.I.T. not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  M.I.T. makes
 * no representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * THIS SOFTWARE IS PROVIDED BY M.I.T. ``AS IS''.  M.I.T. DISCLAIMS
 * ALL EXPRESS OR IMPLIED WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT
 * SHALL M.I.T. BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */
#include "opt_compat.h"
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <sys/time.h>

#include <net/if.h>
#include <net/if_mib.h>


#ifdef COMPAT_IA32
#define PTRIN(v)	(void *)(uintptr_t) (v)
#define PTROUT(v)	(u_int32_t)((long)v & 0xFFFFFFFF) 
#endif

extern struct timeval gStackLastChange;

/*
 * A sysctl(3) MIB for generic interface information.  This information
 * is exported in the net.link.generic branch, which has the following
 * structure:
 *
 * net.link.generic	.system			- system-wide control variables
 *						  and statistics (node)
 *			.ifdata.<ifindex>.general
 *						- what's in `struct ifdata'
 *						  plus some other info
 *			.ifdata.<ifindex>.linkspecific
 *						- a link-type-specific data
 *						  structure (as might be used
 *						  by an SNMP agent
 *
 * Perhaps someday we will make addresses accessible via this interface
 * as well (then there will be four such...).  The reason that the
 * index comes before the last element in the name is because it
 * seems more orthogonal that way, particularly with the possibility
 * of other per-interface data living down here as well (e.g., integrated
 * services stuff).
 */

SYSCTL_DECL(_net_link_generic);
SYSCTL_NODE(_net_link_generic, IFMIB_SYSTEM, system, CTLFLAG_RW, 0,
	    "Variables global to all interfaces");
SYSCTL_INT(_net_link_generic_system, IFMIB_IFCOUNT, ifcount, CTLFLAG_RD,
	   &if_index, 0, "Number of configured interfaces");
#ifdef COMPAT_IA32
static inline void tv_64_to_32(struct timeval32 *dst, const struct timeval *src)
{
	dst->tv_sec = src->tv_sec;
	dst->tv_usec = src->tv_usec;
}
static inline void ifd_64_to_32(struct if_data32 *dst, const struct if_data *src)
{
#define COPY(fld) dst->fld = src->fld
	COPY(ifi_type);
	COPY(ifi_physical);
	COPY(ifi_addrlen);
	COPY(ifi_hdrlen);
	COPY(ifi_link_state);
	COPY(ifi_recvquota);
	COPY(ifi_xmitquota);
	COPY(ifi_datalen);
	COPY(ifi_mtu);
	COPY(ifi_metric);
	COPY(ifi_baudrate);
	COPY(ifi_ipackets);
	COPY(ifi_ierrors);
	COPY(ifi_opackets);
	COPY(ifi_oerrors);
	COPY(ifi_collisions);
	COPY(ifi_ibytes);
	COPY(ifi_obytes);
	COPY(ifi_imcasts);
	COPY(ifi_omcasts);
	COPY(ifi_iqdrops);
	COPY(ifi_noproto);
	COPY(ifi_hwassist);
	COPY(ifi_epoch);
	tv_64_to_32(&dst->ifi_lastchange, &src->ifi_lastchange);
	dst->ifi_svmib = PTROUT(src->ifi_svmib);
#undef COPY
}
static inline void ifxd_64_to_32(struct ifx_data32 *dst, const struct ifx_data *src)
{
#define COPY(fld) dst->fld = src->fld
	COPY(ifi_ipackets);
	COPY(ifi_opackets);
	COPY(ifi_ibytes);
	COPY(ifi_obytes);
	COPY(ifi_imcasts);
	COPY(ifi_omcasts);
	COPY(ifi_ibcasts);
	COPY(ifi_obcasts);
	memcpy(dst->ifi_alias, src->ifi_alias,IF_ALIAS_LENGTH+1);
	COPY(ifi_linkTrapEnable);
#undef COPY
}
static inline void rmon_64_to_32(struct rmon_data32 *dst, const struct rmon_data *src)
{
#define COPY(fld) dst->fld = src->fld
	COPY(etherStatsOctets);
	COPY(etherStatsDropEvents);
	COPY(etherStatsCRCAlignErrors);
	COPY(etherStatsUndersizePkts);
	COPY(etherStatsOversizePkts);
	COPY(etherStatsFragments);
	COPY(etherStatsJabbers);
	COPY(etherStatsCollisions);
	COPY(etherStatsPkts64Octets);
	COPY(etherStatsPkts65to127Octets);
	COPY(etherStatsPkts128to255Octets);
	COPY(etherStatsPkts256to511Octets);
	COPY(etherStatsPkts512to1023Octets);
	COPY(etherStatsPkts1024to1518Octets);
	COPY(etherStatsdot3StatsSymbolErrors);
#undef COPY
}
static inline void ifxmd_64_to_32(struct ifxmibdata32 *dst, const struct ifxmibdata *src)
{
#define COPY(fld) dst->fld = src->fld
		strlcpy(dst->ifmd_name, src->ifmd_name, sizeof(src->ifmd_name));
		COPY(ifmd_pcount);
		COPY(ifmd_flags);
		ifd_64_to_32(&dst->ifmd_data, &src->ifmd_data);
		tv_64_to_32(&dst->ifmd_lastchange, &src->ifmd_lastchange);
		ifxd_64_to_32(&dst->ifxmd_data, &src->ifxmd_data);
		rmon_64_to_32(&dst->rmonmd_data, &src->rmonmd_data); 
		COPY(hw_stats);
		COPY(ifmd_snd_len);
		COPY(ifmd_snd_maxlen);
		COPY(ifmd_snd_drops);
#undef COPY
}
#endif
static int
sysctl_ifdata(SYSCTL_HANDLER_ARGS) /* XXX bad syntax! */
{
	int *name = (int *)arg1;
	int error;
	u_int namelen = arg2;
	struct ifnet *ifp;
	struct ifmibdata ifmd;
	size_t dlen;
	char *dbuf;
	struct ifxmibdata ifxmd;
#ifdef COMPAT_IA32
	struct ifxmibdata32 ifxmd32;
	int compat = 0;
	if(req->oldlen == sizeof(struct ifxmibdata32))
		compat = 1;
#endif
	if (namelen != 2)
		return EINVAL;

	if (name[0] <= 0 || name[0] > if_index ||
	    ifaddr_byindex(name[0]) == NULL)
		return ENOENT;

	ifp = ifaddr_byindex(name[0])->ifa_ifp;

	switch(name[1]) {
	default:
		return ENOENT;
	case IFDATA_GENERAL64BIT:
		strlcpy(ifxmd.ifmd_name, ifp->if_xname, sizeof(ifxmd.ifmd_name));

#define COPY(fld) ifxmd.ifmd_##fld = ifp->if_##fld
		COPY(pcount);
		COPY(flags);
		COPY(data);
#undef COPY
		ifxmd.ifmd_lastchange = gStackLastChange;
		ifxmd.ifxmd_data = ifp->if_data.ifi_svmib->ifmib;
		ifxmd.rmonmd_data = ifp->if_data.ifi_svmib->rmon; 
		ifxmd.hw_stats = ifp->if_data.ifi_svmib->hw_stats;
		ifxmd.ifmd_snd_len = ifp->if_snd.ifq_len;
		ifxmd.ifmd_snd_maxlen = ifp->if_snd.ifq_maxlen;
		ifxmd.ifmd_snd_drops = ifp->if_snd.ifq_drops;
#ifdef COMPAT_IA32
		if(compat)
		{
			/* NOTE: This only for the values copied above, atm*/
			ifxmd_64_to_32(&ifxmd32, &ifxmd);
			error = SYSCTL_OUT(req, &ifxmd32, sizeof ifxmd32);
		}
		else
#endif
		error = SYSCTL_OUT(req, &ifxmd, sizeof ifxmd);
		if (error || !req->newptr)
			return error;

		error = SYSCTL_IN(req, &ifxmd, sizeof ifxmd);
		if (error)
			return error;

#define COPY(fld) ifp->if_##fld = ifmd.ifmd_##fld
		COPY(data);
#undef COPY
		ifp->if_data.ifi_svmib->ifmib = ifxmd.ifxmd_data;
                ifp->if_data.ifi_svmib->rmon = ifxmd.rmonmd_data;
	
		ifp->if_snd.ifq_maxlen = ifxmd.ifmd_snd_maxlen;
		ifp->if_snd.ifq_drops = ifxmd.ifmd_snd_drops;
		break;

	case IFDATA_GENERAL:
		bzero(&ifmd, sizeof(ifmd));
		strlcpy(ifmd.ifmd_name, ifp->if_xname, sizeof(ifmd.ifmd_name));

#define COPY(fld) ifmd.ifmd_##fld = ifp->if_##fld
		COPY(pcount);
		COPY(data);
#undef COPY
		ifmd.ifmd_flags = ifp->if_flags | ifp->if_drv_flags;
		ifmd.ifmd_snd_len = ifp->if_snd.ifq_len;
		ifmd.ifmd_snd_maxlen = ifp->if_snd.ifq_maxlen;
		ifmd.ifmd_snd_drops = ifp->if_snd.ifq_drops;

		error = SYSCTL_OUT(req, &ifmd, sizeof ifmd);
		if (error || !req->newptr)
			return error;

		error = SYSCTL_IN(req, &ifmd, sizeof ifmd);
		if (error)
			return error;

#define DONTCOPY(fld) ifmd.ifmd_data.ifi_##fld = ifp->if_data.ifi_##fld
		DONTCOPY(type);
		DONTCOPY(physical);
		DONTCOPY(addrlen);
		DONTCOPY(hdrlen);
		DONTCOPY(mtu);
		DONTCOPY(metric);
		DONTCOPY(baudrate);
#undef DONTCOPY
#define COPY(fld) ifp->if_##fld = ifmd.ifmd_##fld
		COPY(data);
		ifp->if_snd.ifq_maxlen = ifmd.ifmd_snd_maxlen;
		ifp->if_snd.ifq_drops = ifmd.ifmd_snd_drops;
#undef COPY
		break;

	case IFDATA_LINKSPECIFIC:
		error = SYSCTL_OUT(req, ifp->if_linkmib, ifp->if_linkmiblen);
		if (error || !req->newptr)
			return error;

		error = SYSCTL_IN(req, ifp->if_linkmib, ifp->if_linkmiblen);
		if (error)
			return error;

	case IFDATA_DRIVERNAME:
		/* 20 is enough for 64bit ints */
		dlen = strlen(ifp->if_dname) + 20 + 1;
		if ((dbuf = malloc(dlen, M_TEMP, M_NOWAIT)) == NULL)
			return (ENOMEM);
		if (ifp->if_dunit == IF_DUNIT_NONE)
			strcpy(dbuf, ifp->if_dname);
		else
			sprintf(dbuf, "%s%d", ifp->if_dname, ifp->if_dunit);

		error = SYSCTL_OUT(req, dbuf, strlen(dbuf) + 1);
		if (error == 0 && req->newptr != NULL)
			error = EPERM;
		free(dbuf, M_TEMP);
		return (error);
		
	}
	return 0;
}

SYSCTL_NODE(_net_link_generic, IFMIB_IFDATA, ifdata, CTLFLAG_RW,
	    sysctl_ifdata, "Interface table");

