/*-
 * Copyright (c) 2002 Sandvine, Inc.
 * All rights reserved.
 *
 * sv_enable -- change to a new group
 */

#include <sys/cdefs.h>
#include <sys/types.h>

#include <syslog.h>
#include <err.h>
#include <errno.h>
#include <grp.h>
#include <libgen.h>
#include <paths.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "sv_group.h"

static void	 addgroup(const char *grpname);
static void	 doshell(void);
static void	 usage(void);

static struct passwd *pwd;
static uid_t euid;

/* Manipulate effective user ID. */
#define PRIV_START do {				\
		if (seteuid(euid) < 0)		\
			err(1, "seteuid");	\
	} while (0)
#define PRIV_END do {				\
		if (seteuid(getuid()) < 0)	\
			err(1, "seteuid");	\
	} while (0)

int
main(int argc, char *argv[])
{
    	int ch, test = 0;
	char *switchto = 0;
        struct group *gr;

        if ((pwd = getpwuid(getuid())) == NULL)
	{
		errx(1, "unknown user");
	}
	if ((gr = getgrgid(getgid())) == NULL)
	{
		errx(1, "unknown group");
	}
	if (!getuid())
	{
	    errx(1, "you have maximum privileges");
	}
	euid = geteuid();

	/*
	 * Lower effective ID to real UID since suid bit is set...
	 * From now on, if we want the suid bit activated, we need
	 * to encompass the code with PRIV_START and PRIV_END.
	 */
	if (seteuid(getuid()) < 0)
	{
		err(1, "seteuid");
	}

	while ((ch = getopt(argc, argv, "-n")) != -1) {
		switch (ch) {
		case 'n':
			test = 1;
			break;
		default:
			usage();
		}
        }
        argc -= optind;
        argv += optind;

	if (argc == 1)
	{
		switchto = argv[0];
	}
	else if (argc == 0)
	{
		if (strcmp(gr->gr_name, "sv_operator") == 0)
		{
			switchto = "sv_service";
		}
		else if (strcmp(gr->gr_name, "sv_service") == 0)
		{
			switchto = "sv_admin";
		}
	}

	if (switchto)
	{
		addgroup(switchto);
		if (!test)
			doshell();
	}
	else
	{
		usage();
	}

	return 0;
}

static void
usage(void)
{

	fprintf(stderr, "usage: sv_enable [group]\n");
	exit(1);
}

static void 
addgroup(const char *grpname)
{
    gid_t grp_id;
    long lgid;
    int dbmember, dbfound, i, fd;
    struct group *grp;
    char *ep;
    char **p;
    char groups [CMD_BUF_SIZE+1];

    /* Try it as a group name, then a group id. */
    if ((grp = getgrnam(grpname)) == NULL)
    {
	if ((lgid = strtol(grpname, &ep, 10)) <= 0 || *ep != '\0' ||
	    (grp = getgrgid((gid_t)lgid)) == NULL )
	{
	    errx(1, "%s: bad group name", grpname);
	}
    }

    /*
     * If the user is not a member of the requested group, then check
     * the "allowable groups" list.  If that doesn't work,  simply
     * deny access.
     */
    dbmember = 0;
    if (pwd->pw_gid == grp->gr_gid)
    {
	dbmember = 1;
    }
    for (p = grp->gr_mem; *p != NULL; p++)
    {
	if (strcmp(*p, pwd->pw_name) == 0)
	{
	    dbmember = 1;
	    break;
	}
    }
    if (!dbmember)
    {
	/* Read the "allowable groups" file and see if group is in there.
	 * Only the SANDVINE_GROUPS_FILE_LOCAL is guaranteed to exist.
	 * Hence, the reason for no error reporting on non-local file. */

	dbfound = 0;
	PRIV_START;

	fd = open(SANDVINE_GROUPS_FILE_LOCAL, O_RDONLY|O_CREAT|O_SHLOCK, 0600);
	if (fd != -1)
	{
	    if (!SearchSvGroupsEntry(fd, pwd->pw_name, groups))
	    {
		dbfound = 1;
		if (strstr(groups,grpname))
		{
		    dbmember = 1;
		}
	    }
	    if (close(fd) == -1)
	    {
		errx(1, "Could not close %s.", SANDVINE_GROUPS_FILE_LOCAL);
	    }
	}

	if (!dbfound)
	{
	    fd = open(SANDVINE_GROUPS_FILE, O_RDONLY|O_CREAT|O_SHLOCK, 0600);
	    if (fd != -1)
	    {
		if (!SearchSvGroupsEntry(fd, pwd->pw_name, groups))
		{
		    dbfound = 1;
		    if (strstr(groups,grpname))
		    {
			dbmember = 1;
		    }
		}
		if (close(fd) == -1)
		{
		    errx(1, "Could not close %s.", SANDVINE_GROUPS_FILE);
		}
	    }
	    else
	    {
		errx(1, "Could not open %s.", SANDVINE_GROUPS_FILE);
	    }
	}

	PRIV_END;

	if (!dbmember)
	{
	    errx(1, "Sorry, you are not a member of %s.", grpname);
	}
    }

    /*
     * If the user is trying to lower privilege, then deny access.
     */
    if (grp->gr_gid == getgid())
    {
	errx(1, "You already have that privilege.");
    }
    else if (grp->gr_gid > getgid())
    {
	errx(1, "You may not lower your privilege.");
    }

    /* Set the current process's GID to the one requested by this
     * command.
     */
    PRIV_START;
    if (setgid(grp->gr_gid))
    {
	PRIV_END;
	errx(1, "setgid");
    }
    /* Set the group list of the current process to the one requested
     * by this command.
     */
    grp_id = grp->gr_gid;
    if (setgroups(1, (const gid_t *)&grp_id) < 0)
    {
	PRIV_END;
	errx(1, "setgroups");
    }
    /* If the group we are trying to join is sv_admin, then we need
     * to set the UID of the current process to root's UID.
     */
    if (!strcmp(grp->gr_name, SANDVINE_SUPERUSER_GROUP))
    {
	setuid(0);
	return;
    }
    PRIV_END;
}

/*
 * Invoke a new shell with the current *uid's.
 */
static void
doshell(void)
{
        const char *shell;

        shell = pwd->pw_shell;
        if (*shell == '\0')
        {
                shell = _PATH_BSHELL;
        }
        if (!strcmp(shell, "/bin/bash"))
        {
                execl(shell, basename(shell), "--login", (char *)NULL);
        }
        else if (!strcmp(shell, "/bin/sh"))
        {
                execl(shell, "-i", (char *)NULL);
        }
        else
        {
                execl(shell, basename(shell), (char *)NULL);
        }
        err(1, "%s", shell);
}
